/*
 * Copyright 2011 Christos KK Loverdos
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *      http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

package com.ckkloverdos.convert

import select.ConverterSelectionStrategy
import com.ckkloverdos.maybe._

/**
 * An immutable registry for converters.
 *
 * @author Christos KK Loverdos <loverdos@gmail.com>.
 */
class Converters(selector: ConverterSelectionStrategy) extends CanConvert {
  def canConvertType(sm: Converter.AnyManifest, tm: Converter.AnyManifest) = {
    selector.find(sm, tm).isJust
  }

  def findConverter[S, T](sm: Manifest[S], tm: Manifest[T]): Maybe[Converter[S, T]] = {
    selector.find(sm, tm).asInstanceOf[Maybe[Converter[S, T]]]
  }

  /**
   * Converts a value or throws an exception if the value cannot be converted.
   */
  @throws(classOf[ConverterException])
  def convertValueEx[S: Manifest, T](sourceValue: S, tm: Manifest[T]): T = {
    val sm = manifest[S]
    findConverter(sm, tm) match {
      case Just(cv) =>
        cv.convertEx(sourceValue)
      case NoVal =>
        ConverterException("Could not find converter from %s -> %s for value %s", sm, tm, sourceValue)
      case Failed(exception, explanation) =>
        ConverterException(exception, "Error [%s] trying to find converter from %s -> %s for value %s", explanation, sm, tm, sourceValue)
    }
  }

  def convertValue[S: Manifest, T](sourceValue: S, tm: Manifest[T]): Maybe[T] = {
    findConverter(manifest[S], tm) flatMap (_.convert(sourceValue))
  }

  def convertValueToByte[S: Manifest](sourceValue: S): Maybe[Byte] = convertValue(sourceValue, Manifest.Byte)
  def convertValueToShort[S: Manifest](sourceValue: S): Maybe[Short] = convertValue(sourceValue, Manifest.Short)
  def convertValueToBoolean[S: Manifest](sourceValue: S): Maybe[Boolean] = convertValue(sourceValue, Manifest.Boolean)
  def convertValueToInt[S: Manifest](sourceValue: S): Maybe[Int] = convertValue(sourceValue, Manifest.Int)
  def convertValueToLong[S: Manifest](sourceValue: S): Maybe[Long] = convertValue(sourceValue, Manifest.Long)
  def convertValueToFloat[S: Manifest](sourceValue: S): Maybe[Float] = convertValue(sourceValue, Manifest.Float)
  def convertValueToDouble[S: Manifest](sourceValue: S): Maybe[Double] = convertValue(sourceValue, Manifest.Double)
}

object Converters {
  val DefaultConverters = new StdConvertersBuilder().registerDefaultConversions().build
}